/* ######################################################################### */
/* ##                                                                     ## */
/* ##  S T A C K T H R U D                                                ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Job .......: Server part of a program to send data from host A     ## */
/* ##               B without processing them (with the goal to test for  ## */
/* ##               maximum thruput the stack allows).                    ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Copyright (C) 2002  Daniel Scheibli                                ## */
/* ##                                                                     ## */
/* ##  This program is free software; you can redistribute it and/or      ## */
/* ##  modify it under the terms of the GNU General Public License        ## */
/* ##  as published by the Free Software Foundation; either version 2     ## */
/* ##  of the License, or (at your option) any later version.             ## */
/* ##                                                                     ## */
/* ##  This program is distributed in the hope that it will be useful,    ## */
/* ##  but WITHOUT ANY WARRANTY; without even the implied warranty of     ## */
/* ##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ## */
/* ##  GNU General Public License for more details.                       ## */
/* ##                                                                     ## */
/* ##  You should have received a copy of the GNU General Public License  ## */
/* ##  along with this program; if not, write to the Free Software        ## */
/* ##  Foundation, Inc., 59 Temple Place - Suite 330, Boston,             ## */
/* ##  MA  02111-1307, USA.                                               ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Author ....: Daniel Scheibli <daniel@scheibli.com>                 ## */
/* ##  Date ......: 2002-03-05                                            ## */
/* ##  Changes ...: ----- stackthrud V00.99.006 ------------------------- ## */
/* ##               2006-12-02 o Update code layout and program output.   ## */
/* ##               ----- stackthrud V00.99.005 ------------------------- ## */
/* ##               2004-06-18 o Added concurrent connection support.     ## */
/* ##                          o Deactivated UDP part.                    ## */
/* ##                          o Applied new code layout.                 ## */
/* ##               ----- stackthrud V00.99.004 ------------------------- ## */
/* ##               <none>                                                ## */
/* ##               ----- stackthrud V00.99.003 ------------------------- ## */
/* ##               <none>                                                ## */
/* ##               ----- stackthrud V00.99.002 ------------------------- ## */
/* ##               <none>                                                ## */
/* ##               ----- stackthrud V00.99.001 ------------------------- ## */
/* ##               <none>                                                ## */
/* ##                                                                     ## */
/* ######################################################################### */



// Presettings

#include<stdlib.h>
#include<stdio.h>
#include<errno.h>
#include<sys/types.h>
#include<sys/socket.h>
#include<netinet/in.h>

#define TRUE      ( 1 == 1 )
#define FALSE     ( 1 == 2 )

#define PRINTF    printf
#define STRICMP   strcasecmp

#define EOS       '\0'
#define NEWLINE   "\n"



// Globale variables

char *pcParamListenTo;
char *pcParamCommunication; 
int   iParamBlockSize;



// Function prototypes

int Communication_TcpV4( void );
int Communication_UdpV4( void );





/* ######################################################################### */
/* ##                                                                     ## */
/* ##   < M A I N >                                                       ## */
/* ##                                                                     ## */
/* ######################################################################### */
int main( int argc, char *argv[] ) {

	// Presettings

	int iRc;



	// Process parameters

	PRINTF("%s", NEWLINE);
	PRINTF( "STACKTHRUD V00.99.006%s", NEWLINE );
	PRINTF( "%s", NEWLINE );

	if( argc != ( 1 + 3 ) ) {
		PRINTF( "Usage ........: stackthrud >>--- LISTENTO --- COMMUNICATION --->      %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "                            >--- BLOCKSIZE ---><                      %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "Parameters ...: LISTENTO      = Port,... where the stackthru-daemon   %s", NEWLINE );
		PRINTF( "                                is listening:                         %s", NEWLINE );
		PRINTF( "                                - For TCP/IP family: <port>           %s", NEWLINE );  
		PRINTF( "                COMMUNICATION = The protocol/way which should be used:%s", NEWLINE );
		PRINTF( "                                - TCPv4 = TCP Version 4               %s", NEWLINE );  
		PRINTF( "                                - UDPv4 = UDP Version 4               %s", NEWLINE );
		PRINTF( "                BLOCKSIZE     = Number of bytes per read() from the   %s", NEWLINE );
		PRINTF( "                                stackthru client.                     %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "Sample .......: stackthrud 2342 TCPv4 1448                            %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "                Listens to port 2342 for TCP packages and reads       %s", NEWLINE );
		PRINTF( "                in 1448 byte blocks.                                  %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		exit( 1 );
	}
	else {
		pcParamListenTo      = argv[1];
		pcParamCommunication = argv[2];  
		iParamBlockSize      = atoi( argv[3] );
	}



	// Call the selected communication method

	iRc = -1;

	if( STRICMP( pcParamCommunication, "TCPV4" ) == 0 ) { 
		iRc = Communication_TcpV4( );
	}
	if( STRICMP( pcParamCommunication, "UDPV4" ) == 0 ) { 
		PRINTF( "===> ERROR: Not implemented.%s", NEWLINE );
		exit( 1 ); 
	}

	if( iRc != 0 ) {
		PRINTF( "===> ERROR: The selected communication method does not exists%s", NEWLINE );
		PRINTF( "            or returned with an errror (RC: %i)%s", iRc, NEWLINE );
		exit( 1 );
	}



	// Return program status

	return( 0 );
}





/* ######################################################################### */
/* ##                                                                     ## */
/* ##   C O M M U N I C A T I O N _ T C P V 4                             ## */
/* ##                                                                     ## */
/* ######################################################################### */
int Communication_TcpV4( void ) {

	// Presettings

	int                 iParamPort      = -1;
	char               *pcFrameData     = NULL;
	int                 iSocketFd;
	int                 iClientSocketFd;
	struct sockaddr_in  sInetAddr;
	int                 iDummy;
	int                 iRc;



	// Seperate convert and check port string

	iParamPort = atoi( pcParamListenTo );
	if( iParamPort <= 0 ) {
		PRINTF( "===> ERROR: The format of the given port is invalid!%s", NEWLINE );
		return( 1 );
	}



	// Build buffer for frame data to be read

	pcFrameData = malloc( iParamBlockSize );
	if( pcFrameData == NULL ) {
		PRINTF( "===> ERROR: Allocation of memory for frame data failed!%s", NEWLINE );
		return( 1 );
	}
	memset( pcFrameData, EOS, iParamBlockSize );



	// Create and configure socket file descriptor
	
	iSocketFd = socket( AF_INET, SOCK_STREAM, 0 );
	if( iSocketFd < 0 ) {
		PRINTF( "===> ERROR: Failed while creating socket!%s", NEWLINE );
		return( 1 );
	}
	iDummy = 1;
	iRc = setsockopt( iSocketFd, SOL_SOCKET, SO_REUSEADDR, &iDummy, sizeof( iDummy ) );
	if( iRc != 0 ) {
		PRINTF( "===> ERROR: Failed while configuring socket!%s", NEWLINE );
		PRINTF( "     errno: %i%s", errno, NEWLINE );
		return( 1 );
	}



	// Try to bind socket

	sInetAddr.sin_family      = AF_INET;
	sInetAddr.sin_port        = htons( iParamPort );
	sInetAddr.sin_addr.s_addr = INADDR_ANY;   // 0.0.0.0 = 0x00000000

	iRc = bind( iSocketFd, ( struct sockaddr * ) &sInetAddr, sizeof( sInetAddr ) );
	if( iRc < 0 ) {
		PRINTF( "===> ERROR: Failed while binding socket!%s", NEWLINE );
		PRINTF( "     errno: %i%s", errno, NEWLINE );
		return( 1 );
	}



	// Try to listen to socket
	iRc = listen( iSocketFd, 5 );
	if( iRc != 0 ) {
		PRINTF( "===> ERROR: Failed while listening to port!%s", NEWLINE );
		PRINTF( "     errno: %i%s", errno, NEWLINE );
		return( 1 );
	}



	// Loop over incoming connections

	while( TRUE ) {
	
		// Wait for incoming connection

		iDummy = sizeof( sInetAddr );
		iClientSocketFd = accept( iSocketFd, ( struct sockaddr * ) &sInetAddr, &iDummy );
		if( iClientSocketFd < 0 ) {
			PRINTF( "===> ERROR: Failed while waiting for clients!%s", NEWLINE );
			PRINTF( "     errno: %i%s", errno, NEWLINE );
			return( 1 );
		}
		PRINTF( "Session established ...%s", NEWLINE );

		// Fork out to process the clients request

		switch( fork( ) ) {
			case -1:
				PRINTF( "===> ERROR: Failed while forking to serve a client request!%s", NEWLINE );
				PRINTF( "     errno: %i%s", errno, NEWLINE );
				exit( 1 );
			case 0:
				// Close listen socket

				close( iSocketFd );
				
				// Read from socket/client

				for( iRc = 42; iRc > 0; ) {
					iRc = read( iClientSocketFd, pcFrameData, iParamBlockSize );
				}
				if( iRc < 0 ) {
					PRINTF( "===> ERROR: Failed while reading from client!%s", NEWLINE );
					PRINTF( "     errno: %i%s", errno, NEWLINE );
					exit( 1 );
				}				

				// Close connection

				close( iClientSocketFd );				
				PRINTF( "Session terminated ...%s", NEWLINE );
				exit( 0 );
			default:
				// Close the new accepted socket
				
				close( iClientSocketFd );
				break;
		}
	}



	// End function
	
	return( 0 );
}




